/*
 * mpgedit curses interface
 *
 * Copyright (C) 2001 Adam Bernstein
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307  USA.
 */

#ifndef lint
static char SccsId[] = "$Id: mpegcurses.c,v 1.28 2001/10/03 14:21:02 number6 Exp $";
#endif

#include <curses.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include "mp3time.h"
#include "mp3_header.h"
#include "mpegcurses.h"


#ifndef MAXPATHLEN
#define MAXPATHLEN 1024
#endif

#ifndef KEY_ESC
#define KEY_ESC 27
#endif

typedef struct _terminal_chars {
    int *values;
    int len;
} terminal_chars;

typedef struct _input_field {
    int    cols;
    int    value;
    int    curxpos;
    char   sep;
    int    zeropad;
    int    maxval;
    struct _input_field *rnext;
    struct _input_field *lnext;
} input_field;

typedef struct _input_fname {
    char name[MAXPATHLEN];
    int len;
} input_fname;


typedef struct _input_time {
    input_field min;
    input_field sec;
    input_field msec;
    struct _input_field *cursor;
} input_time;


typedef struct _edit_line {
    input_time stime;
    input_time etime;
    input_time *cursor;
    input_fname filename;
} edit_line;


typedef struct _edit_array {
    edit_line **lines;
    edit_line *cursor;
    int       max;
    int       alloc_max;
    int       i;
    int       start;
} edit_array;


typedef struct _delete_line_array {
    edit_line *lines[20];
    int i;
    int max;
} delete_line_array;

typedef struct _edit_window {
    WINDOW *win;
    int y;
    int x;
    int maxy;
    int maxx;
} edit_window;


char *curs_time2string(input_time *t, char *str);
char *curs_edit_line_format(edit_window *ewin, edit_line *e, char *buf);
void curs_edit_line_write(edit_window *ewin, char *buf);


int find_terminal(terminal_chars *t, int c)
{
    int done = 0;
    int i;

    for (i=0; i < t->len; i++) {
        if (t->values[i] == c) {
            done = 1;
            break;
        }
    }
    return done;
}


void curs_time_fixup(input_time *t)
{
    if (t->msec.value < 0) {
        t->msec.value = t->msec.maxval-1;
        t->sec.value--;
    }
    if (t->sec.value < 0) {
        t->sec.value = t->sec.maxval-1;
        t->min.value--;
    }
    if (t->min.value < 0) {
        t->min.value = 0;
        t->min.value = t->min.maxval-1;
    }

    if (t->msec.value >= t->msec.maxval) {
        t->msec.value = 0;
        t->sec.value++;
    }
    if (t->sec.value >= t->sec.maxval) {
        t->sec.value = 0;
        t->min.value++;
    }
    if (t->min.value >= t->min.maxval) {
        t->min.value = 0;
    }
}


void curs_clear_time(input_time *t)
{
    t->min.value = 0;
    t->sec.value = 0;
    t->msec.value = 0;
}


int curs_input_time_init(input_time *t, int line, int col)
{
    int status = 1;

    memset(t, 0, sizeof(*t));

    t->min.cols  = 3;
    t->sec.cols  = 2;
    t->msec.cols = 3;

    t->min.curxpos  = col;
    t->sec.curxpos  = col + 3;
    t->msec.curxpos = col + 7;

    t->min.sep  = ':';
    t->sec.sep  = '.';
    t->msec.sep = ' ';

    t->min.maxval = 1000;
    t->sec.maxval = 60;
    t->msec.maxval = 1000;

    t->min.zeropad = 0;
    t->sec.zeropad = 1;
    t->msec.zeropad = 1;

    t->min.rnext  = &t->sec;
    t->sec.rnext  = &t->msec;
    t->msec.rnext = &t->min;

    t->min.lnext  = &t->msec;
    t->sec.lnext  = &t->min;
    t->msec.lnext = &t->sec;

    curs_time_fixup(t);
    return status;
}


int curs_input_fname_init(input_fname *f, int width)
{
    int status = 1;

    memset(f, 0, sizeof(*f));
    f->len = width;

    return status;
}


void curs_delete_line_array_init(delete_line_array *d)
{
    memset(d, 0, sizeof(*d));
    d->max = sizeof(d->lines)/sizeof(delete_line_array *);
}


edit_line *curs_edit_line_undelete(delete_line_array *d)
{
    edit_line *l = NULL;

    if (d->i > 0) {
        d->i--;
    }
    l = d->lines[d->i];
    d->lines[d->i] = NULL;
    return l;
}


void curs_edit_line_delete(edit_line *e, delete_line_array *d)
{
    int i;

    if (d->i >= d->max) {
        free(d->lines[0]);
        d->i--;
        for (i=0; i < d->i; i++) {
            d->lines[i] = d->lines[i+1];
        }
    }
    d->lines[d->i++] = e;
}


void curs_edit_array_display(edit_window *ewin, edit_array *earray)
{
    int winy, indx;

    /* TBD: compute width based upon window width */
    char fmtbuf[1024];

    wmove(ewin->win, 0, 0);
    wclrtobot(ewin->win);
    for (winy=0, indx=earray->start; indx<=earray->max; winy++, indx++) {
        mvwprintw(ewin->win, winy, 0, 
                  curs_edit_line_format(ewin, earray->lines[indx], fmtbuf));
    }
    wmove(ewin->win, ewin->y, 0);
    curs_edit_line_write(ewin,
                         curs_edit_line_format(ewin,
                                               earray->cursor,
                                               fmtbuf));
    wrefresh(ewin->win);
}


/*
 * Format element values into canonical string format for 
 * display on screen.
 */
char *curs_edit_line_format(edit_window *ewin, edit_line *e, char *buf)
{
    char *cp;
    char *sp;
    int  len;
    int  pad;

    cp = buf;
    len = strlen(e->filename.name);
    /*
     * TBD: change filename.len to filename.width, as this is the
     * screen display width.
     */
    if (len > e->filename.len) {
        sp = e->filename.name + (len - e->filename.len);
    }
    else {
        sp = e->filename.name;
    }
    strcpy(cp, sp);

    pad = ewin->maxx - strlen(cp) - 25 /* 2 times time field width */;
    while (pad > 0) {
        strcat(cp, " ");
        pad--;
    }
    cp = cp + strlen(cp);

    curs_time_fixup(&e->stime);
    curs_time2string(&e->stime, cp);
    strcat(cp, " ");

    cp = cp + strlen(cp);
    curs_time_fixup(&e->etime);
    curs_time2string(&e->etime, cp);

    return buf;
}


void curs_edit_line_write(edit_window *ewin, char *buf)
{
    wmove(ewin->win, ewin->y, ewin->x);
    wclrtoeol(ewin->win);
    wattron(ewin->win, A_REVERSE);
    mvwprintw(ewin->win, ewin->y, ewin->x,  buf);
    wattroff(ewin->win, A_REVERSE);
    wrefresh(ewin->win);
}



int curs_edit_window_init(edit_window *ewin, int winy, int winx)
{

    int status = 1;

    memset(ewin, 0, sizeof(*ewin));
    ewin->win = newwin(winy-5, winx, 1, 0);
    if (!ewin->win) {
        status = 0;
    }
    getmaxyx(ewin->win, ewin->maxy, ewin->maxx);
    keypad(ewin->win,    TRUE);
    scrollok(ewin->win,  FALSE);
    return status;
}


edit_line *curs_edit_line_init(int line, int col, int winx)
{
    int status = 1;
    int fname_width;
    edit_line *e;

    e = (edit_line *) calloc(1, sizeof(edit_line));
    if (!e) {
        return NULL;
    }

    fname_width = winx - 25;
    status = curs_input_fname_init(&e->filename, fname_width);
    if (!status) {
        goto clean_exit;
    }

    status = curs_input_time_init(&e->stime, line, winx-23);
    if (!status) {
        goto clean_exit;
    }
    e->cursor = &e->stime;

    status = curs_input_time_init(&e->etime, line, winx-12);
    if (!status) {
        goto clean_exit;
    }

    if (!status) {
        free(e);
        e = NULL;
    }
clean_exit:
    return e;
}


void curs_write_str(WINDOW *win, char *str, int width)
{
    char *cp;
    int len;


    len = strlen(str);
    if (len >= width) {
        cp = str + (len - width);
    }
    else {
        cp = str;
    }
    wmove(win, 0, 0);
    wclrtoeol(win);
    wprintw(win, cp);
    wrefresh(win);
}



void curs_set_fname(input_fname *f, char *name)
{
    strcpy(f->name, name);
}


int curs_read_str(WINDOW *win,
                  char *str,
                  int ypos,
                  int xpos,
                  int win_width,
                  terminal_chars *t,
                  int *last)
{
    int done = 0;
    int len = 0;
    int refresh = 0;
    int input;
    int xcursor = 0;
    int width;
    char *cp;
    int previous = 0;

    width = win_width - xpos;
    len = strlen(str);
    if (len > 0) {
        refresh = 1;
    }
    do {
        if (refresh) {
            if (len >= width) {
                cp = str + (len - width);
            }
            else {
                cp = str;
            }
            mvwprintw(win, ypos, xpos, "%-*s", width, cp);
            wmove(win, ypos, xcursor);
            wrefresh(win);
        }
        refresh = 0;
        xcursor = xpos + len;
        if (len >= width) {
            xcursor = xpos + width - 1;
        }
        else {
            xcursor = xpos + len;
        }
        input = mvwgetch(win, ypos, xcursor);
        if (isprint(input)) {
            refresh = 1;
            if (input == '\\') {
                if (previous == input) {
                    str[len++] = input;
                    previous   = 0;
                }
                else {
                    previous = input;
                }
            }
            else {
                str[len++] = input;
                previous = input;
            }
            str[len] = '\0';
        }
        else if (input == KEY_BACKSPACE) {
            if (len == 0) {
                str[len] = '\0';
            }
            else {
                str[--len] = '\0';
            }
            refresh = 1;
        }
        else if (input == KEY_LEFT) {
            str[0]  = '\0';
            len     = 0;
            refresh = 1;
        }

        done = find_terminal(t, input);
        if (done && input != '\n') {
            len = strlen(str);
            refresh = 1;
        }
    } while (!done);
    *last = input;

    /*
     * Trim white space from start and end of string
     */
    refresh = 0;
    for (cp = str; isspace(*cp); cp++) 
        ;
    if (cp != str) {
        memmove(str, cp, strlen(cp));
        refresh = 1;
    }
    str[strlen(cp)] = '\0';
    for (cp = str+strlen(str)-1; cp != str && isspace(*cp); cp--) {
        *cp = '\0';
        refresh = 1;
    }
    if (refresh) {
        curs_write_str(win, str, width);
    }
    return *last;
}


int curs_read_integer(edit_window *ewin,
                      edit_line *el,
                      input_field *inf,
                      terminal_chars *t,
                      int *last)
{
    char str[8];
    int  len  = 0;
    int  done = 0;
    int  input;
    int  refresh;

    /* TBD: compute width based upon window width */
    char fmtbuf[1024];

    str[0] = '\0';
    do {
        refresh = 0;
        input = mvwgetch(ewin->win, ewin->y, inf->curxpos);
        if (input < UCHAR_MAX && isdigit(input)) {
            if (len < inf->cols) {
                str[len++] = input;
                str[len]   = '\0';
            }
            else {
                len = 0;
                str[len++] = input;
                str[len]   = '\0';
            }
            if (atoi(str) >= inf->maxval) {
                len = 0;
                str[len]   = '\0';
            }
            refresh = 1;
        }
        else if (input == KEY_BACKSPACE) {
            if (len == 0) {
                str[len] = '\0';
            }
            else {
                str[--len] = '\0';
            }
            refresh = 1;
        }

        if (refresh) {
            inf->value = atoi(str);
            curs_edit_line_format(ewin, el, fmtbuf);
            curs_edit_line_write(ewin, fmtbuf);
        }
        done = find_terminal(t, input);
    } while (!done);
    *last = input;
    return(strlen(str) == 0  ? -1 : atoi(str));
}


int curs_read_fname(WINDOW *win, int y, int x, input_fname *f, char *prompt)
{
    int last;
    int done = 0;
    terminal_chars int_terminal;
    int int_term_array[] = { '\n' , KEY_RIGHT, KEY_ESC };
    int xoffset = 0;

    
    wattron(win, A_REVERSE);
    mvwprintw(win, y, x, "%*s", f->len, " ");

    if (prompt) {
        mvwprintw(win, y, x, prompt);
        wmove(win, y, strlen(prompt));
        xoffset = strlen(prompt);
    }
    wrefresh(win);
    int_terminal.values = int_term_array;
    int_terminal.len    = sizeof(int_term_array)/sizeof(int);

    last = curs_read_str(win, f->name,
                         y, xoffset, f->len, &int_terminal, &last);
    done = 1;
    wattroff(win, A_REVERSE);
    return last;
}


int curs_read_time(edit_window *ewin, edit_line *el, input_time *intime)
{
    int            value = 0;
    int            done = 0;
    int            last;
    input_field    *t;

    /* TBD: compute width based upon window width */
    char fmtbuf[1024];

    terminal_chars int_terminal;
    int            int_term_array[] = 
        { 'f', 'n', ' ', KEY_LEFT, KEY_RIGHT, KEY_UP, KEY_DOWN,
          ',', '.', 'h', 'l', 'k', 'j', 'c', '\t',
          's', 'S', 'e', 'E', 'q', 'Q', 'L', 'o', 'O', 'A', 'p', 'P', 'D',
          'H', 'G', KEY_ESC };

    int_terminal.values = int_term_array;
    int_terminal.len    = sizeof(int_term_array)/sizeof(int);

    if (intime->cursor) {
        t = intime->cursor;
    }
    else {
        t = &intime->min;
    }

    do {
        value = curs_read_integer(ewin,
                                  el,
                                  t,
                                  &int_terminal, &last);
        
        if (value != -1) {
            t->value = value % t->maxval;
            curs_edit_line_format(ewin, el, fmtbuf);
            curs_edit_line_write(ewin, fmtbuf);
        }

        switch(last) {
          case '.':
            t->value++;
            curs_edit_line_format(ewin, el, fmtbuf);
            curs_edit_line_write(ewin, fmtbuf);
            break;

          case ',':
            t->value--;
            curs_edit_line_format(ewin, el, fmtbuf);
            curs_edit_line_write(ewin, fmtbuf);
            break;

          case KEY_RIGHT:
          case 'l':
          case ' ':
            if (t == &intime->msec) {
                done = 1;
            }
            else {
                t = t->rnext;
            }
            break;

          case KEY_LEFT:
          case 'h':
            if (t == &intime->min) {
                done = 1;
            }
            else {
                t = t->lnext;
            }
            break;

          case 'c':
            curs_clear_time(intime);
            t->value               = 0;
            t->rnext->value        = 0;
            t->rnext->rnext->value = 0;
            t                      = &intime->min;
            intime->cursor         = t;
            curs_edit_line_format(ewin, el, fmtbuf);
            curs_edit_line_write(ewin, fmtbuf);
            break;

          case KEY_UP:
          case KEY_DOWN:
          default:
            done = 1;
            break;
        }
    } while (!done);
    intime->cursor = t;
    return last;
}


char *curs_time2string(input_time *t, char *str)
{
    sprintf(str, "%3d:%.2d.%.3d", t->min.value, t->sec.value, t->msec.value);
    return str;
}


void curs_string2time(char *str, input_time *t)
{
    char *cp;
    int val;

    val = strtol(str, &cp, 10); 
    t->min.value = val;
    cp++;

    val = strtol(cp, &cp, 10); 
    t->sec.value = val;
    cp++;
    
    val = strtol(cp, &cp, 10); 
    t->msec.value = val;
    cp++;
}


int curs_input_time_iszero(input_time *t) 
{
    return t->min.value == 0 && t->sec.value == 0 && t->msec.value == 0;
}



/*
 * Returns 0  if equal,
 *         <0 if t1 < t2
 *         >0 if t1 > t2
 *        -1 if t1 or t2 is NULL.
 * Nothing can be assumed by the numeric value returned, as it may
 * be the difference between seconds or milliseconds. 
 */
int curs_input_time_cmp(input_time *t1, input_time *t2)
{
    int sec1;
    int sec2;
    int msec1;
    int msec2;
    
    /* Not exactly correct, but must return something for this case */
    if (!t1 || !t2) {
        return -1;
    }

    sec1  = t1->min.value * 60 + t1->sec.value;
    sec2  = t2->min.value * 60 + t2->sec.value;
    msec1 = t1->msec.value;
    msec2 = t2->msec.value;

    /* Equality test */
    if (sec1 == sec2 && msec1 == msec2) {
        return 0;
    }

    /* Must differ only by milliseconds */
    if (sec1 == sec2) {
        return msec1-msec2;
    }

    /* Must differ by seconds */
    return sec1-sec2;
}


char *curs_build_edit_cmdstring(edit_array *earray)
{
    int len = 0;
    int i;
    char *cp;
    char str[16];
    int maxlen;
    int iszero1;
    int iszero2;


    maxlen = earray->max+1;
    /* 32 is 2 * string size of a time value */
    len = maxlen * (sizeof(earray->lines[0]->filename.name) + 32);
    cp = (char *) calloc(1, len);
    if (!cp) {
        return cp;
    }
    for (i=0; i<maxlen; i++) {
        iszero1 = curs_input_time_iszero(&earray->lines[i]->stime);
        iszero2 = curs_input_time_iszero(&earray->lines[i]->etime);
        if (earray->lines[i]->filename.name && (!iszero1 || !iszero2)) {
            strcat(cp, " -e");
            if (!iszero1) {
                curs_time2string(&earray->lines[i]->stime, str);
                strcat(cp, str);
            }
            strcat(cp, "-");
            if (!iszero2) {
                curs_time2string(&earray->lines[i]->etime, str);
                strcat(cp, str);
            }
            strcat(cp, " -f ");
            strcat(cp, earray->lines[i]->filename.name);
        }
    }
    if (!cp[0]) {
        free(cp);
        cp = NULL;
    }
    return cp;
}


editspec *curs_build_editspec(edit_array *earray, int *cnt, int *status)
{
    int i;
    int j = 0;
    int maxlen;
    editspec *edits;
    char str[12];
    char timespec[32];
    int iszero1;
    int iszero2;
    int error = 0;


    maxlen = earray->max+1;
    edits = (editspec *) calloc(maxlen, sizeof(editspec));
    if (!edits) {
        return edits;
    }

    for (i=0; i<maxlen; i++) {
        iszero1 = curs_input_time_iszero(&earray->lines[i]->stime);
        iszero2 = curs_input_time_iszero(&earray->lines[i]->etime);
        timespec[0] = '\0';
        if (earray->lines[i]->filename.name &&
            *earray->lines[i]->filename.name)
        {
            /*
             * Both values cannot be zero.
             */
            if (iszero1 && iszero2) {
                error = 2;
            }
            else {
                /*
                 * Test the constraint that time2 must be greater than time1
                 */
                if (!iszero1 && !iszero2 &&
                    curs_input_time_cmp(&earray->lines[i]->etime,
                                        &earray->lines[i]->stime) <= 0)
                {
                        error = 1;
                }
                if (!iszero1) {
                    curs_time2string(&earray->lines[i]->stime, str);
                    strcat(timespec, str);
                }
                strcat(timespec, "-");
                if (!iszero2) {
                    curs_time2string(&earray->lines[i]->etime, str);
                    strcat(timespec, str);
                }
                edits[j].filename = strdup(earray->lines[i]->filename.name);
                edits[j].timespec = strdup(timespec);
                j++;
            }
        }
        else {
            /*
             * Flag an error when no filename is filled in but there 
             * is edit specification.
             */ 
            error = 3;
        }
    }
    if (error) {
        free(edits);
        edits = NULL;
    }
    else {
        *cnt = j;
    }
    if (status) {
        *status = error;
    }
    return edits;
}



int ttyio_getch(void *ctx)
{
    return getch();
}


static void ttyio_printf(void *ctx, const char *fmt, ...)
{
    /* TBD: compute width based upon window width */
    char buf[1024];
    WINDOW *win = (WINDOW *) ctx;

    va_list ap;
    va_start(ap, fmt);
    vsprintf(buf, fmt, ap);
    wprintw(win, buf);
    wrefresh(win);
    va_end(ap);
}


int curs_call_indexfile(char *file, WINDOW *win)
{

    int                 status = 0;
    mpegfio_iocallbacks ttyio;

    memset(&ttyio, 0, sizeof(ttyio));

    ttyio.getch  = ttyio_getch;
    ttyio.printf = ttyio_printf;
    ttyio.ctx    = win;

    /*
     * Move cursor to the output box area, just in case
     * the index file will be created.  Text put occurs.
     */
    wmove(win, 0, 0);
    wclrtoeol(win);
    wmove(win, 1, 0);
    if (curs_build_indexfile(file, &ttyio)) {
        wmove(win, 0, 0);
        wclrtobot(win);

        mvwprintw(win, 2, 0,  "Failed opening file '%s'",
                  file);
        file[0] = '\0';
        status = 1;
    }
    else {
        wmove(win, 0, 0);
        wclrtobot(win);
    }
    wrefresh(win);
    return status;
}



int curs_load_abandoned_edits(char        *file,
                              edit_window *ewin,
                              edit_array  *earray,
                              WINDOW      *output,
                              int         winx)
{
    FILE      *fp;
    int       i;
    char      inbuf[MAXPATHLEN+32];
    char      *namep;
    char      *stp;
    char      *etp;
    char      *cp;
    edit_line **realloc_lines;
    int       sts = 1;

    /* TBD: compute width based upon window width */
    char fmtbuf[1024];

    fp = fopen(file, "r");
    if (!fp) {
        return 1;
    }

    
    i = 0;
    do {
        cp = fgets(inbuf, sizeof(inbuf)-1, fp);
        if (!cp) {
            /* Most likely EOF */
            continue;
        }
        
        inbuf[strlen(inbuf)-1] = '\0';
        namep = stp = etp = NULL;
        cp = strstr(inbuf, "%%");
        if (!cp) {
            goto clean_exit;
        }
        if (cp) {
            namep = inbuf;
            *cp = '\0';
            cp+=2;
            stp = cp;
        }
        if (cp < inbuf+sizeof(inbuf) && *cp) {
            cp = strstr(cp, "%%");
            if (cp) {
                *cp = '\0';
                cp+=2;
                etp = cp;
            }
        }
        if (namep && stp && etp) {
            earray->i = i;
            if (i > earray->alloc_max) {
                realloc_lines = (edit_line **)
                    realloc(earray->lines,
                            sizeof(edit_line *) * earray->alloc_max * 2);
                if (!realloc_lines) {

                    /* Not enough memory for operation */
                    /* adam error message here */

                    mvwprintw(output, 2, 0,
                              "Failed allocating more memory for edit buffer");
                    wrefresh(output);
                    break;
                }
                else {
                    earray->lines = realloc_lines;
                    earray->alloc_max *= 2;
                }
            }
            if (i > earray->max) {
                earray->lines[earray->i] = 
                    curs_edit_line_init(1+earray->i, 0, winx);
                if (!earray->lines[earray->i]) {
                     fprintf(stderr, "%s: %s\n",
                             "curs_load_abandoned_edits",
                             "curs_edit_line_init failed");
                     return 1;
                }
                earray->max = earray->i;
            }
            strcpy(earray->lines[i]->filename.name, namep);
            curs_string2time(stp, &earray->lines[i]->stime);
            curs_string2time(etp, &earray->lines[i]->etime);
            curs_set_fname(&earray->lines[i]->filename, namep);
            curs_call_indexfile(namep, output);

            curs_edit_line_format(ewin, earray->lines[i], fmtbuf);
            i++;
            ewin->y++;
        }
    } while (!feof(fp));
    sts = 0;
    ewin->y--;

    /*
     * Position viewport on first screen of loaded data when
     * more lines were loaded than can be displayed.
     */
    if (ewin->y >= ewin->maxy) {
        ewin->y = ewin->maxy - 1;
        earray->i = ewin->y;
    }
    earray->cursor = earray->lines[earray->i];
    curs_edit_array_display(ewin, earray);
    wmove(ewin->win, ewin->y, ewin->x);
    wrefresh(ewin->win);

clean_exit:
    fclose(fp);
    return sts;
}


int curs_save_abandoned_edits(edit_array *earray, char *filename)
{
    FILE *fp;
    int i;
    char stbuf[32];
    char etbuf[32];
    char *cp;

    fp = fopen(filename, "w");
    if (!fp) {
        return 1;
    }
    for (i=0; i<=earray->max; i++) {
        cp = earray->lines[i]->filename.name;
        curs_time2string(&earray->lines[i]->stime, stbuf);
        curs_time2string(&earray->lines[i]->etime, etbuf);
        fprintf(fp, "%s%%%%%s%%%%%s\n", cp, stbuf, etbuf);
    }
    fclose(fp);
    return 0;
}


int curs_get_out_filename(input_fname *filename,
                          char *prompt,
                          char *errprompt,
                          WINDOW *win,
                          int uniq)
{
    int status;
    int inchar;
    input_fname response;

    curs_input_fname_init(&response, 70);
    do {
        inchar = curs_read_fname(win, 2, 0,
                                 filename, prompt);
        if (inchar == KEY_ESC) {
            status = 0;
            break;
        }
        status = validate_outfile(filename->name, NULL, NULL);
        if (status) {
            wmove(win, 0, 0);
            wclrtobot(win);
            if (status == 1) {
                flash();
                if (!uniq) {
                    strcpy(response.name, "");
                    inchar = curs_read_fname(win, 2, 0, &response, errprompt);
                    if (!response.name[0]) {
                        strcpy(response.name, "No");
                    }
                    if (response.name[0] == 'y' || response.name[0] == 'Y') {
                        status = 0;
                    }
                    else if (response.name[0] == 'n' || response.name[0] == 'N') {
                        inchar = KEY_ESC;
                        status = 0;
                    }
                    else if (inchar == KEY_ESC) {
                        status = 0;
                        break;
                    }
                }
                else {
                    mvwprintw(win, 1, 0,
                              "Output file must be unique");
                    wrefresh(win);
                }
            }
            else if (status == 2) {
                mvwprintw(win, 2, 0,
                          "Append file is not writable '%s'",
                          filename->name);
            }
            else {
                mvwprintw(win, 2, 0,
                          "Unable to use output file '%s'",
                          filename->name);
            }
        }
        else {
            wmove(win, 0, 0);
            wclrtobot(win);
        }
    } while (status);
    return inchar;
}



int curs_get_abandoned_filename(input_fname *filename, WINDOW *win)
{
    int status = 0;
    int inchar;
    int done = 0;

    do {
        inchar = curs_read_fname(win, 1, 0,
                                 filename, "Abandoned file: ");
        done = inchar == '\n' || inchar == KEY_ESC || inchar == KEY_RIGHT;
        status = access(filename->name, F_OK);
        if (status == -1) {
            wmove(win, 0, 0);
            wclrtoeol(win);
            wmove(win, 2, 0);
            wclrtoeol(win);
            mvwprintw(win, 2, 0,
                      "File does not exist '%s'",
                       filename->name);
        }
        else {
            wmove(win, 0, 0);
            wclrtobot(win);
        }
        wrefresh(win);
    } while (status == -1 && !done);
    return inchar;
}


char *curs_get_default_abandoned_filename(char *name)
{
    FILE  *fp;

    fp = fopen(".mp3edit_abandonded_name", "r");
    if (!fp) {
        return NULL;
    }
    fgets(name, MAXPATHLEN, fp);
    fclose(fp);
    return name;
}


int curs_put_default_abandoned_filename(char *name)
{
    FILE *fp;
    int sts;

    fp = fopen(".mp3edit_abandonded_name", "w");
    if (!fp) {
        return 1;
    }
    sts = fputs(name, fp);
    fclose(fp);
    return sts;
}


int curs_play(char *filename,
              mpegfio_iocallbacks *ttyio,
              editspec **ret_edits,
              int *ret_edits_len,
              int *outfile_flag,
              char **ret_out_filename)
{
    edit_array          earray;
    edit_window         ewin;
    input_fname         out_filename;
    input_fname         abandoned_filename;
    input_fname         quit_response;
    input_time          *intime;
    input_time          *lasttime = NULL;
    input_time          tmp_stime;
    input_fname         save_fname;
    edit_line           save_editline;
    WINDOW              *output;
    WINDOW              *rootwin;
    int                 inchar;
    int                 quit = 0;
    int                 status;
    char                stime_fmt[32];
    char                etime_fmt[32];
    char                duration_time_fmt[32];
    char                *cmdstr;
    editspec            *edits = NULL;
    int                 edits_len;
    int                 winy, winx;
    int                 i;
    delete_line_array   undos;
    edit_line           *undo_line;
    edit_line           **realloc_lines;
    int                 already_loaded = 0;
    int                 sts;

    /* TBD: compute width based upon window width */
    char fmtbuf[1024];

    rootwin = initscr();
    if (!rootwin) {
        fprintf(stderr, "curs_play: Curses initialization initscr() failed\n");
        return 1;
    }
    noecho();
    raw();

    /* Get size of screen for later use */
    getmaxyx(rootwin, winy, winx);

    /* Place title information on the root window; this shows through */
    mvprintw(0, 0, "     Input file");
    mvprintw(0, winx-25, "Start time   End time");
    refresh();

    stime_fmt[0] = '\0';
    etime_fmt[0] = '\0';

    status = curs_edit_window_init(&ewin, winy, winx);
    if (!status) {
        fprintf(stderr, "curs_edit_window_init failed\n");
        return 1;
    }

    curs_delete_line_array_init(&undos);
    memset(&earray, 0, sizeof(earray));

    /* Allocate buffer based upon window size */

    earray.alloc_max = winy * 2;
    earray.lines = (edit_line **) calloc(earray.alloc_max, sizeof(edit_line *));
    if (!earray.lines) {
        fprintf(stderr, "failed allocating initial edit_line array\n");
        return 1;
    }
    earray.lines[0] = curs_edit_line_init(1, 0, winx);
    if (!earray.lines[0]) {
        fprintf(stderr, "curs_edit_line_init(eline) failed\n");
        return 1;
    }

    earray.cursor = earray.lines[0];

    status = curs_input_fname_init(&out_filename, winx-1);
    if (!status) {
        fprintf(stderr, "curs_input_fname_init(out_filename) failed\n");
        return 1;
    }

    status = curs_input_fname_init(&abandoned_filename, winx-1);
    if (!status) {
        fprintf(stderr, "curs_input_fname_init(abandoned_filename) failed\n");
        return 1;
    }
    status = curs_input_fname_init(&quit_response, winx-1);
    if (!status) {
        fprintf(stderr, "curs_input_fname_init(quit_response) failed\n");
        return 1;
    }

    output = newwin(3, winx, winy-3, 0);
    if (!output) {
        fprintf(stderr, "newwin(output) failed\n");
        return 1;
    }
    scrollok(output, TRUE);
    keypad(output, TRUE);

    /*
     * I/O Context for playback function. 
     */
    ttyio->getch  = ttyio_getch;
    ttyio->printf = ttyio_printf;
    ttyio->ctx    = output;

    intime = earray.cursor->cursor;
    if (filename) {
        curs_call_indexfile(filename, output);
        curs_set_fname(&earray.lines[0]->filename, filename);
    }
    else {
        curs_set_fname(&earray.lines[0]->filename, "-input MP3 file-");
    }
    curs_edit_line_format(&ewin, earray.cursor, fmtbuf);
    curs_edit_line_write(&ewin, fmtbuf);
    save_editline = *earray.cursor;
    do {
        inchar = curs_read_time(&ewin, earray.lines[earray.i], intime);
        wmove(output, 0, 0);
        wclrtobot(output);
        switch (inchar) {

          /* Restore current edit line when ESC is pressed */
          case KEY_ESC:
            *earray.cursor = save_editline;
            curs_edit_line_write(&ewin,
                                 curs_edit_line_format(&ewin,
                                                       earray.cursor,
                                                       fmtbuf));
            break;

          /* Move cursor one edit line up */
          case KEY_UP:
          case 'k':
            earray.i--;
            if (earray.i < 0) {
                earray.i = 0;
            }

            ewin.y--;
            if (ewin.y < 0) {
                ewin.y = 0;
                earray.start--;
                if (earray.start < 0) {
                    earray.start = 0;
                }
            }
            earray.cursor = earray.lines[earray.i];
            intime = earray.cursor->cursor;
            curs_edit_array_display(&ewin, &earray);
            save_editline = *earray.cursor;
            break;
    
          /* Move cursor down */

          case KEY_DOWN:
          case 'j':
            if (earray.i < earray.max) {
                earray.i++;
                if ((ewin.y+1) < ewin.maxy) {
                    ewin.y++;
                }
                else {
                    earray.start++;
                }
                earray.cursor = earray.lines[earray.i];
            }
            intime = earray.cursor->cursor;
            curs_edit_array_display(&ewin, &earray);
            save_editline = *earray.cursor;
            break;

          /* Paste previously deleted line, above cursor line */

          case 'P':
            undo_line = curs_edit_line_undelete(&undos);
            if (!undo_line) {
                break;
            }

            /* Move all entries starting at cursor down one line */

            for (i=earray.max; i>=earray.i; i--) {
                earray.lines[i+1] = earray.lines[i];
            }
            earray.lines[earray.i] = undo_line;
            earray.max++;
            earray.cursor = earray.lines[earray.i];
            intime        = earray.cursor->cursor;
            curs_edit_array_display(&ewin, &earray);
            save_editline = *earray.cursor;
            break;

          /* Paste previously deleted line, below cursor line */

          case 'p':
            undo_line = curs_edit_line_undelete(&undos);
            if (!undo_line) {
                break;
            }

            /* Move all entries starting at cursor down one line */

            for (i=earray.max; i>=earray.i; i--) {
                earray.lines[i+1] = earray.lines[i];
            }
            earray.i++;
            if ((ewin.y+1) < ewin.maxy) {
                ewin.y++;
            }
            else {
                earray.start++;
            }
            earray.lines[earray.i] = undo_line;
            earray.max++;
            earray.cursor = earray.lines[earray.i];
            intime        = earray.cursor->cursor;
            curs_edit_array_display(&ewin, &earray);
            save_editline = *earray.cursor;
            break;

          /* Add line below current cursor position */

          case 'o':
            /*
             * Allocate more space for edit array
             */
            if (earray.max + 1 >= earray.alloc_max) {
                realloc_lines = (edit_line **)
                    realloc(earray.lines,
                            sizeof(edit_line *) * earray.alloc_max * 2);
                if (!realloc_lines) {

                    /* Not enough memory for operation */

                    mvwprintw(output, 2, 0,
                              "Failed allocating more memory for edit buffer");
                    break;
                }
                else {
                    earray.lines = realloc_lines;
                    earray.alloc_max *= 2;
                }
            }

            /*
             * Shift viewport in edit array down one line when the
             * cursor is at the bottom of the screen.
             */
            if ((ewin.y+1) >= ewin.maxy) {
                earray.start++;
            }
            else {
                ewin.y++;
            }


            /* Move all entries starting at cursor down one line */

            for (i=earray.max; i>=earray.i; i--) {
                earray.lines[i+1] = earray.lines[i];
            }
            earray.i++;

            earray.lines[earray.i] = curs_edit_line_init(1+earray.i, 0, winx);
            if (!earray.lines[earray.i]) {
                fprintf(stderr, "curs_edit_line_init(eline) failed\n");
                return 1;
            }
            earray.max++;
            earray.cursor = earray.lines[earray.i];
            intime        = earray.cursor->cursor;

            if (earray.i > 0) {
                /*
                 * Initialize new line filename to previous line's name
                 */
                curs_set_fname(&earray.lines[earray.i]->filename,
                               earray.lines[earray.i-1]->filename.name);
                /*
                 * Initialize the stime to the previous fields etime
                 */
                earray.lines[earray.i]->stime.min.value =
                    earray.lines[earray.i-1]->etime.min.value;
                earray.lines[earray.i]->stime.sec.value =
                    earray.lines[earray.i-1]->etime.sec.value;
                earray.lines[earray.i]->stime.msec.value =
                    earray.lines[earray.i-1]->etime.msec.value;
            }

            curs_edit_line_format(&ewin, earray.cursor, fmtbuf);
            curs_edit_array_display(&ewin, &earray);
            save_editline = *earray.cursor;

            break;

          /* Add line above current cursor position */

          case 'O':
            /*
             * Allocate more space for edit array
             */
            if (earray.max + 1 >= earray.alloc_max) {
                realloc_lines = (edit_line **)
                    realloc(earray.lines,
                            sizeof(edit_line *) * earray.alloc_max * 2);
                if (!realloc_lines) {

                    /* Not enough memory for operation */

                    mvwprintw(output, 2, 0,
                              "Failed allocating more memory for edit buffer");
                    break;
                }
                else {
                    earray.lines = realloc_lines;
                    earray.alloc_max *= 2;
                }
            }

            /* Move all entries starting at cursor down one line */

            for (i=earray.max; i>=earray.i; i--) {
                earray.lines[i+1] = earray.lines[i];
            }
            earray.lines[earray.i] = curs_edit_line_init(1+earray.i, 0, winx);
            if (!earray.lines[earray.i]) {
                fprintf(stderr, "curs_edit_line_init(eline) failed\n");
                return 1;
            }
            earray.max++;
            earray.cursor = earray.lines[earray.i];
            intime        = earray.cursor->cursor;

            if (earray.i > 0) {
                /*
                 * Initialize new line filename to previous line's name
                 */
                curs_set_fname(&earray.lines[earray.i]->filename,
                               earray.lines[earray.i-1]->filename.name);
                /*
                 * Initialize the stime to the previous fields etime
                 */
                earray.lines[earray.i]->stime.min.value =
                    earray.lines[earray.i-1]->etime.min.value;
                earray.lines[earray.i]->stime.sec.value =
                    earray.lines[earray.i-1]->etime.sec.value;
                earray.lines[earray.i]->stime.msec.value =
                    earray.lines[earray.i-1]->etime.msec.value;
            }
            else {
                curs_set_fname(&earray.lines[0]->filename,
                               "-input MP3 file-");
            }
            curs_edit_line_format(&ewin, earray.cursor, fmtbuf);
            curs_edit_array_display(&ewin, &earray);
            save_editline = *earray.cursor;
            break;
    
          case 'D':
            /*
             * Delete the edit line at the cursor position.
             */
            if (earray.max > 0) {
                if (earray.cursor == earray.lines[earray.max]) {
                    /*
                     * Cursor is positioned at the bottom of the list.
                     * Erase the line from the screen, then decrement the
                     * list size by one.
                     */
                    curs_edit_line_delete(earray.lines[earray.i], &undos);
                    earray.i--;
                    ewin.y--;
                    earray.max--;
                    earray.cursor = earray.lines[earray.i];
                    intime = earray.cursor->cursor;
                    curs_edit_array_display(&ewin, &earray);
                }
                else {
                    curs_edit_line_delete(earray.lines[earray.i], &undos);

                    /*
                     * Must copy everything below the current cursor position
                     * up in the array.  This is where a linked list would
                     * be much nicer.
                     */
                    for (i=earray.i; i<earray.max; i++) {
                        earray.lines[i] = earray.lines[i+1];
                    }

                    earray.max--;
                    earray.cursor = earray.lines[earray.i];
                    intime = earray.cursor->cursor;
                    curs_edit_array_display(&ewin, &earray);
                }
                save_editline = *earray.cursor;
            }
            break;

          case 'n':
            inchar = curs_get_out_filename(&out_filename,
                                           "Output file: ",
                                           "Output file exists, overwrite? ",
                                           output, 1);
            if (inchar == KEY_ESC) {
                out_filename.name[0] = '\0';
            }
            break;
    
          case 'f':
            save_fname = earray.cursor->filename;
            inchar = curs_read_fname(ewin.win, ewin.y, ewin.x, 
                                     &earray.cursor->filename, NULL);

            status = curs_call_indexfile(earray.cursor->filename.name, output);
            if (status) {
                earray.cursor->filename = save_fname;
            }
            else {
                earray.lines[earray.i]->stime.min.value  = 0;
                earray.lines[earray.i]->stime.sec.value  = 0;
                earray.lines[earray.i]->stime.msec.value = 0;
            }
            curs_edit_line_write(&ewin, 
                  curs_edit_line_format(&ewin, earray.cursor, fmtbuf));
            break;
    
          /* Toggle between stime/etime, moving cursor to right */
          case '\t':
          case 'l':
          case KEY_RIGHT:
          case ' ':
            if (intime == &earray.cursor->stime) {
                earray.cursor->etime.cursor = &earray.cursor->etime.min;
                earray.cursor->cursor       = &earray.cursor->etime;
            }
            else {
                earray.cursor->stime.cursor = &earray.cursor->stime.min;
                earray.cursor->cursor       = &earray.cursor->stime;
            }
            intime = earray.cursor->cursor;
            break;

          /* Toggle between stime/etime, moving cursor to left */
          case 'h':
          case KEY_LEFT:
            if (intime == &earray.cursor->stime) {
                earray.cursor->etime.cursor = &earray.cursor->etime.msec;
                earray.cursor->cursor       = &earray.cursor->etime;
            }
            else {
                earray.cursor->stime.cursor = &earray.cursor->stime.msec;
                earray.cursor->cursor       = &earray.cursor->stime;
            }
            intime = earray.cursor->cursor;
            break;
    
          /* Sound playback beginning at the current time field */

          case 's':
          case 'S':
            if (earray.cursor->filename.name &&
                earray.cursor->filename.name[0]) 
            {
                if (lasttime) {
                    intime = lasttime;
                    lasttime = NULL;
                }
                wmove(output, 2, 0);
                wclrtobot(output);
                wrefresh(output);
                nodelay(rootwin, TRUE);

                memset(&tmp_stime, 0, sizeof(tmp_stime));
                if (inchar == 's') {
                    strcpy(duration_time_fmt, "1000000:00.0");
                    if (intime == &earray.cursor->stime) {
                        tmp_stime = earray.cursor->stime;
                    }
                    else {
                        tmp_stime = earray.cursor->etime;
                    }
                    curs_time2string(&tmp_stime, stime_fmt);
                }
                else {
                    strcpy(duration_time_fmt, "0:05.0");
                    if (intime == &earray.cursor->stime) {
                        tmp_stime = earray.cursor->stime;
                    }
                    else {
                        tmp_stime = earray.cursor->etime;
                    }
                    tmp_stime.sec.value -= 5;
                    if (tmp_stime.sec.value < 0) {
                        tmp_stime.sec.value += 60;
                        tmp_stime.min.value -= 1;
                    }
                    if (tmp_stime.min.value < 0) {
                        tmp_stime.min.value = tmp_stime.sec.value = 0;
                    }
                    curs_time2string(&tmp_stime, stime_fmt);
                }
                
                status = 
                    curs_interactive_play_file(earray.cursor->filename.name,
                                               stime_fmt,
                                               duration_time_fmt, ttyio);

                nodelay(rootwin, FALSE);
                if (status) {
                    wmove(output, 0, 0);
                    wclrtobot(output);
                    if (status == 1 || status == 4) {
                        mvwprintw(output, 2, 0,
                                  "Can't open index file for '%s'",
                                  earray.cursor->filename.name);
                 
                    }
                    else if (status == 2) {
                        mvwprintw(output, 2, 0,"Can't open play file '%s'",
                                  earray.cursor->filename.name);
                    }
                    else if (status == 3) {
                        mvwprintw(output, 2, 0,
                                  "Can't open external MP3 player");
                    }
                    else if (status == 5) {
                        mvwprintw(output, 2, 0,
                                  "Invalid seek time");
                    }
                    else {
                        mvwprintw(output, 2, 0, "Error playing file '%s'",
                                  earray.cursor->filename.name);
                    }
                }
            }
            else {
                wmove(output, 0, 0);
                wclrtobot(output);
                mvwprintw(output, 2, 0,  "No input file specified");
            }
    
            break;

          /* Move cursor to top of edit array */

          case 'H':
            earray.i      = 0;
            earray.start  = 0;
            ewin.y        = 0;
            earray.cursor = earray.lines[earray.i];
            intime        = earray.cursor->cursor;
            curs_edit_array_display(&ewin, &earray);
            save_editline = *earray.cursor;
            break;

          /* Move cursor to bottom of edit array */

          case 'G':
            earray.i      = earray.max;
            earray.start  = earray.max - ewin.maxy + 1;
            if (earray.start < 0) {
                earray.start = 0;
            }
            ewin.y        = ewin.maxy - 1;
            if (ewin.y > earray.max) {
                ewin.y = earray.max;
            }
            earray.cursor = earray.lines[earray.i];
            intime        = earray.cursor->cursor;
            curs_edit_array_display(&ewin, &earray);
            save_editline = *earray.cursor;
            break;

          /* Load abandoned edits file */
          case 'L':
            if (already_loaded) {
                mvwprintw(output, 2, 0, "Edits have already been loaded");
                wrefresh(output);
                break;
            }

            curs_get_default_abandoned_filename(abandoned_filename.name);
            if (!abandoned_filename.name[0]) {
                strcpy(abandoned_filename.name, ".mp3edit_abandoned");
            }
            inchar = curs_get_abandoned_filename(&abandoned_filename, output);
            if (inchar != KEY_ESC && abandoned_filename.name[0]) {
                sts = curs_load_abandoned_edits(abandoned_filename.name,
                                                &ewin, &earray, output, winx);
                if (sts == 0) {
                    already_loaded = 1;
                    curs_put_default_abandoned_filename(abandoned_filename.name);
                    intime = earray.cursor->cursor;
                    save_editline = *earray.cursor;
                }
                else {
                    mvwprintw(output, 2, 0, "Failed loading abandoned edits file");
                    wrefresh(output);
                }
            }
            break;

          case 'Q':
            strcpy(quit_response.name, "");
            do {
                inchar = 
                    curs_get_out_filename(&quit_response,
                                          "Quit current edits?: ", 
                                          "",
                                          output, 0);
            } while (inchar != '\n');
            if (quit_response.name[0] == 'y' || quit_response.name[0] == 'Y') {
                abandoned_filename.name[0] = '\0';
                quit = -1;
            }
            break;

          case 'q':
            curs_get_default_abandoned_filename(abandoned_filename.name);
            if (!abandoned_filename.name[0]) {
                strcpy(abandoned_filename.name, ".mp3edit_abandoned");
            }
            do {
                inchar = curs_get_out_filename(
                             &abandoned_filename,
                             "Abandoned file: ",
                             "Abandoned edits file exists, overwrite? ",
                             output, 0);
            } while (!abandoned_filename.name[0] && inchar != KEY_ESC);
            if (inchar != KEY_ESC && abandoned_filename.name[0]) {
                quit = -1;
                curs_put_default_abandoned_filename(abandoned_filename.name);
            }
            else {
                wmove(output, 0, 0);
                wclrtobot(output);
            }
            break;

          case 'e':
          case 'E':
            /*
             * 'E' performs same action as 'e', except will write all edits
             * to derived file names, instead of requiring the specification
             * of an output filename.
             */
            if (inchar == 'E') {
                *outfile_flag = 1;
            }

            edits = curs_build_editspec(&earray, &edits_len, &status);
            if (status) {
                wmove(output, 0, 0);
                wclrtobot(output);
                if (status == 3) {
                    mvwprintw(output, 2, 0,
                              "Missing filename in edit specification");
                }
                else {
                    mvwprintw(output, 2, 0,
                              "Error with edit time specified",
                              inchar, inchar, inchar);
                }
                flash();
            }
            else if (!*outfile_flag &&(!out_filename.name ||
                     !*out_filename.name))
            {
                wmove(output, 0, 0);
                wclrtobot(output);
                mvwprintw(output, 2, 0,  "Output filename required",
                          inchar, inchar, inchar);
                flash();
            }
            else {
                curs_get_default_abandoned_filename(abandoned_filename.name);
                if (!abandoned_filename.name[0]) {
                    strcpy(abandoned_filename.name, ".mp3edit_abandoned");
                }
                do {
                    inchar =
                        curs_get_out_filename(
                            &abandoned_filename,
                            "Abandoned file: ", 
                            "Abandoned edits file exists, overwrite? ",
                            output, 0);
                } while (!abandoned_filename.name[0] && inchar != KEY_ESC);
                if (inchar != KEY_ESC && abandoned_filename.name[0]) {
                    wmove(output, 0, 0);
                    wclrtobot(output);
                    curs_put_default_abandoned_filename(
                        abandoned_filename.name);
                    quit = 1;
                }
            }
            if (quit == 0) {
                *outfile_flag = 0;
            }
            break;
           
          default:
            mvwprintw(output, 0, 0,  "Read character %c %d %x",
                      inchar, inchar, inchar);
            break;
        }
        wrefresh(output);
    } while (!quit);

    /* Final refresh clears screen at end */
    clrtobot();
    refresh();

    cmdstr = curs_build_edit_cmdstring(&earray);
    mvprintw(2, 0, "");
    refresh();

    if (ret_edits && ret_edits_len && edits && edits_len > 0) {
        *ret_edits     = edits;
        *ret_edits_len = edits_len;
    }
    if (out_filename.name && *out_filename.name) {
        if (ret_out_filename) {
            *ret_out_filename = strdup(out_filename.name);
        }
    }
    endwin();
    if (abandoned_filename.name[0]) {
        fprintf(stderr, "Saving abandoned edits\n");
        curs_save_abandoned_edits(&earray, abandoned_filename.name);
    }
    return quit < 0;
}
