/*
 * mpgedit file I/O and buffering routines
 *
 * Copyright (C) 2001 Adam Bernstein
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307  USA.
 */

#ifndef lint
static char SccsId[] = "$Id: mpegfio.c,v 1.6 2001/10/20 14:51:38 number6 Exp $";
#endif

#include <stdio.h>
#include <string.h>
#include "mpegfio.h"
#include "header.h"

void mpeg_file_iobuf_clear(mpeg_file_iobuf *buf)
{
    memset(buf, 0, sizeof(*buf));
}


unsigned char *mpeg_file_iobuf_getptr(mpeg_file_iobuf *buf)
{
    return buf->buf + buf->start;
}


void mpeg_file_iobuf_setstart(mpeg_file_iobuf *buf, int start)
{
    buf->start = start;
}

int mpeg_file_iobuf_getstart(mpeg_file_iobuf *buf)
{
    return buf->start;
}


int mpeg_file_iobuf_getlen(mpeg_file_iobuf *buf)
{
    return buf->len - buf->start;
}


int mpeg_file_iobuf_read(FILE *fp, mpeg_file_iobuf *buf)
{
    int sts = 1;

    if (feof(fp)) {
        sts = 0;
        goto clean_exit;
    }
    if (buf->start == 0 && buf->len == 0) {

        /* Fill empty buffer */

        buf->len = fread(buf->buf, 1, sizeof(buf->buf), fp);
        if (buf->len == 0 && feof(fp)) {
            sts = 0;
            goto clean_exit;
        }
    }
    else if (buf->start < buf->len) {
        /*
         * Move remaining data up to start of buffer, then fill in
         * remaining space.
         */
         memmove(buf->buf, buf->buf+buf->start, buf->len-buf->start);
         buf->len  = buf->len - buf->start;
         buf->len += fread(buf->buf + buf->len, 1, buf->start, fp);
         buf->start = 0;
    }
    else {
        sts = 0;
    }


clean_exit:
    return sts;
}


int read_next_mpeg_frame(FILE *fp,
                         mpeg_file_iobuf *mpegiobuf,
                         mpeg_header_data *header,
                         int *eof)
{
    int found   = 0;
    int not_eof = 1;
    int offset  = 0;

    do {
        if (mpeg_file_iobuf_getlen(mpegiobuf) <= 4 || found == -1) {
            /*
             * Not a full MPEG frame header present in the buffer, so read
             * the next buffer's worth of data from the file.  Advance
             * the input buffer by offset, so the residual data is preserved.
             */
            if (offset > 0) {
                mpeg_file_iobuf_setstart(mpegiobuf, offset);
            }

            not_eof = mpeg_file_iobuf_read(fp, mpegiobuf);
        }
        offset = 0;

        /*
         * Search for the next MPEG frame header in the buffer of data.
         */
        if (not_eof) {
            found = find_mpeg_header_buf(mpeg_file_iobuf_getptr(mpegiobuf),
                                         mpeg_file_iobuf_getlen(mpegiobuf),
                                         header, 0, &offset);
    
            if (found == 0) {
                /*
                 * Have not yet found the start of the MPEG frame.   Move the
                 * data buffer cursor forward over data already processed.
                 */
                mpeg_file_iobuf_setstart(mpegiobuf,
                                         mpeg_file_iobuf_getstart(mpegiobuf) +
                                         mpeg_file_iobuf_getlen(mpegiobuf) - 4);
            }
        }
    } while (found != 1 && not_eof);

    *eof = not_eof == 0;

    if (eof && found == -1) {
        /*
         * Because there is no frame following the last frame in a file,
         * found will be -1.
         */
        found = 1;
    }
    if (found && offset > 0) {
        mpeg_file_iobuf_setstart(mpegiobuf,
                                 mpeg_file_iobuf_getstart(mpegiobuf) + offset);
    }
    return found;
}


int mpegfio_has_xing_header(FILE *fp,
                            mpeg_header_data *mpeg_header,
                            XHEADDATA *xingh, int append)
{
    int status = 0;
    int cnt = 0;
    long curpos;
    unsigned char    mpeg_header_data[4];

    
    /*
     * Remember where we are in the file; may need to return here.
     */
    curpos = ftell(fp);

    /*
     * Test to see if this is a VBR file with XING header.
     * When appending to the end of a file, read the header from the
     * target file, and update  the current edits stats in that header.
     */
    memset(xingh, 0, sizeof(*xingh));
    if (append) {
        fseek(fp, 0, SEEK_SET);
    }

    cnt = fread(mpeg_header_data, sizeof(mpeg_header_data), 1, fp);
    if (cnt != 1) {
        goto clean_exit;
    }
    cnt = decode_mpeg_header(mpeg_header_data, mpeg_header, 0);
    if (!cnt) {
        goto clean_exit;
    }
    memcpy(xingh->xingbuf, mpeg_header_data, sizeof(mpeg_header_data));
    cnt = fread(&xingh->xingbuf[4], mpeg_header->frame_size - 4, 1, fp);
    if (!cnt) {
        goto clean_exit;
    }
    if (append) {
        fseek(fp, 0, SEEK_END);
    }

    status = (get_xing_header(xingh->xingbuf, xingh) > 0);
    /*
     * Return to same position in data stream when no header is present 
     */
clean_exit:
    if (!status) {
        fseek(fp, curpos, SEEK_SET);
    }

    return status ? mpeg_header->frame_size : 0;
}
